﻿-- Synchronizing module

-- spinning tiles fading in/out times
local SPINNING_TILE_FADE_IN_TIME = 0.07
local SPINNING_TILE_FADE_OUT_TIME = 0.37
local SPINNING_TILE_SIZE = 27

-- client syncing state
-- careful about changing 0, it is hardcoded in the main module
local SYNCING_STATE_NONE = 0
-- set if waiting for newer versions (upon logging in/getting into party or raid)
local SYNCING_STATE_CHECK_VERSION = 1
-- set if this has performed a new search
local SYNCING_STATE_NEW_SEARCH = 2
-- set if received new search key and started listening
local SYNCING_STATE_RECEIVING_NEW_SEARCH = 3
-- set if sender sent all the items (after own search or logging in) and is waiting for more accurate predictions
local SYNCING_STATE_NARROWING_SEARCH = 4
-- set if receiver is waiting for full item details (after receiving short item details, recalculating new predictions and sending them back)
local SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS = 5
-- set if receiver has just logged in and is waiting for newer searches or better predictions
local SYNCING_STATE_AWAITING_BETTER_SEARCH = 6
BMTSyncingState = SYNCING_STATE_NONE

-- syncing state texts (for spinning inset in the main frame)
local SYNCING_STATE_CHECK_VERSION_TEXT = "Synchronizing Version"
local SYNCING_STATE_NEW_SEARCH_TEXT = "Synchronizing Items - New Search"
local SYNCING_STATE_NARROWING_SEARCH_TEXT = "Attempting to narrow time predictions"
local SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS_TEXT = "Waiting for full item details"
local SYNCING_STATE_AWAITING_BETTER_SEARCH_TEXT = "Checking for more recent search"

-- timeouts (in seconds)
-- after 2 seconds further syncing steps will be taken
local SYNCING_TIMEOUT_VERSION = 2
-- time to send out all the items after starting to send them, otherwise everything will be ignored
local SYNCING_TIMEOUT_NEW_SEARCH = 2
-- same as above, but when requested upon logging in etc.
local SYNCING_TIMEOUT_BETTER_SEARCH_REQ = 3

-- addon message prefix
local ADDON_PREFIX = "BMT"
-- addon message separator (fields)
local SEPARATOR = "%"
-- addon message separator (tables/category)
local SEPARATOR_CATEGORY = "$"
-- syncing keys (strings) sent through addon message
-- version request (also includes this version)
local KEY_VERREQ = "VERREQ"
-- version response
local KEY_VERRESP = "VERRESP"
-- new search key
local KEY_NEW_SEARCH = "NSEA"
-- better predictions resync key
local KEY_BETTER_PREDICTIONS = "BPRED"
-- better search request
local KEY_BETTER_SEARCH_REQ = "BSEAREQ"
-- table for easier parsing (update info)
local updateInfoSyncField = {"playerName",	"playerClass",	"playerGuild",	"updateDate",	"updateTime" }
local isUpdateFieldNumber = {false,			false,			false,			false,			true }
-- short item details key (used for deciding wheather receiver's auction should be moved to history, as well as determining new predictions)
local KEY_ITEM_SHORT = "ISHO"
-- table for easier parsing (update info)
local itemShortSyncField = {"name", "marketID", "timeLeft", "currBid", "numBids", "lastTimeLeftChangeDate", "lastTimeLeftChangeTime", "predictedLow", "predictedHigh" }
local isItemShortFieldNumber = {false, true, true, true, true, false, true, true, true}
-- full item details
local KEY_ITEM_FULL = "IFULL"
local itemFullSyncField = {"name", "marketID", "timeLeft", "currBid", "numBids", "lastTimeLeftChangeDate", "lastTimeLeftChangeTime", "predictedLow", "predictedHigh", "level", "highestBidder", "texture", "itemType", "link", "sellerName", "minIncrement", "quantity", "levelType", "minBid", "predictionsUpdateMinAfterSearch"}
local isItemFullSyncFieldNumber = {false, true, true, true, true, false, true, true, true, true, false, false, false, false, false, true, true, false, true, true}


-- highest revision received from others upon logging in
BMTHighestRevision = BMTRevision;

-- this is a 'copy' of current db for realm received from sync
-- the fields' name should be the same as in the original db
BMTSyncDB = {}

-- this is also a 'copy' of current db for received realm, but 
-- it can store multiple versions of it for the same realm
BMTSyncBetterDB = {}

function BMTSyncing_BroadcastAddonMessage(msg)
	-- guild
	BMT:SendCommMessage(ADDON_PREFIX, msg, "GUILD")
	-- TODO: party/raid (check if in one)
	-- TODO: whispers
end

function BMT:OnCommReceived(prefix, msg, channel, sender)
	if string.len(msg) > 70 then
		BMT:Debug("("..sender..")"..string.sub(msg, 1, 70).."...")
	else
		BMT:Debug("("..sender..")"..msg)
	end
	
	if sender == UnitName("player") then return end
	-- categories (initial info/update info/short items/full items etc.)
	-- initial - sender's addon version and realm being synced
	local categories = {strsplit(SEPARATOR_CATEGORY, msg)}
	if categories == nil then return end
	local initial = {strsplit(SEPARATOR, categories[1])}
	if initial == nil then return end
	local key = initial[1]
	local senderAddonVer = tonumber(initial[2])
	local syncedRealm = initial[3]
	if key == KEY_VERREQ then
		if senderAddonVer > BMTHighestRevision then
			BMTHighestRevision = senderAddonVer
		end
		-- answer using the same channel
		BMT:SendCommMessage(ADDON_PREFIX, BMTSyncing_InitialInfo(KEY_VERRESP), channel, sender)
	elseif key == KEY_VERRESP then
		if senderAddonVer > BMTHighestRevision then
			BMTHighestRevision = senderAddonVer
		end
	elseif key == KEY_NEW_SEARCH and BMTSyncingState == SYNCING_STATE_NONE then
		BMTSyncing_OnNewSearchReceive(categories, syncedRealm, channel, sender)
	elseif key == KEY_BETTER_PREDICTIONS   and (BMTSyncingState == SYNCING_STATE_NEW_SEARCH or BMTSyncingState == SYNCING_STATE_AWAITING_BETTER_SEARCH) then
		BMTSyncing_OnBetterResultReceive(categories, syncedRealm, channel, sender)
	-- always respond to this, regardless of syncing state
	elseif key == KEY_BETTER_SEARCH_REQ then
		BMTSyncng_OnSearchRequestReceive(categories, syncedRealm, channel, sender)
	-- we also write the items if we're doing nothing (for some reason..)
	elseif key == KEY_ITEM_FULL and BMTSyncingState == SYNCING_STATE_NONE or BMTSyncingState == SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS then
		BMTSyncng_OnFullItemDetailsReceive(categories, syncedRealm, channel, sender)
	end
end

-- categories[1] is initial update, 2 is updateInfo, rest is items
function BMTSyncing_OnNewSearchReceive(categories, syncedRealm, channel, sender)
	-- if there was still something, then remove it
	BMTSyncDB[syncedRealm] = nil
	-- create data struct
	BMTSyncDB[syncedRealm] = {}
	BMTSyncDB[syncedRealm]["updateInfo"] = {}
	BMTSyncDB[syncedRealm]["items"] = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	
	-- ignore searches older than our
	local res = BMT_CompareDateAndTime(BMTSyncDB[syncedRealm]["updateInfo"]["updateDate"], BMTSyncDB[syncedRealm]["updateInfo"]["updateTime"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateTime"])
	if res == "smaller" or res == "equal" then BMTSyncDB[syncedRealm] = nil return end
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncDB[syncedRealm]["items"]
		-- (i-2) so we start from 1
		BMTSyncDB[syncedRealm]["items"][(i-2)] = {}
		for j=1, #itemShortSyncField do
			if isItemShortFieldNumber[j] then
				BMTSyncDB[syncedRealm]["items"][(i-2)][itemShortSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncDB[syncedRealm]["items"][(i-2)][itemShortSyncField[j]] = itemInfo[j]
			end
		end
	end
	
	-- rewrite update info to current
	BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"] = BMTSyncDB[syncedRealm]["updateInfo"]
	
	-- here we have everything that will be needed to recalculate better predictions (and if needed, sending them back)
	BMT_MoveAllOldEntriesToHistory(syncedRealm, BMTSyncDB[syncedRealm]["items"], BMTSyncDB[syncedRealm]["updateInfo"]["updateDate"], BMTSyncDB[syncedRealm]["updateInfo"]["updateTime"])
	
	-- here, we only have items that should be on the current list in the BlackMarketTrackerCurrentDB
	-- time to do predictions on our side (treat the search as if it was our own and check if they are better than the received ones)
	local betterPredMsg = BMTSyncing_InitialInfo(KEY_BETTER_PREDICTIONS)
	-- add update info (which is the same as the sender, but we want this message to look the same for both: after search narrowing and after login narrowing)
	for i=1, #updateInfoSyncField do
		if i > 1 then
			betterPredMsg = betterPredMsg .. SEPARATOR
		end
		betterPredMsg = betterPredMsg..BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]]
	end
	betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
	local numItems = BMT_GetNumItems(syncedRealm)
	local currDate = date("%d/%m/%y")
	local currHour, currMinute = GetGameTime()
	local currTime = 60*currHour+currMinute
	for i=1, numItems do
		for j=1, #BMTSyncDB[syncedRealm]["items"] do
			if BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["marketID"] == BMTSyncDB[syncedRealm]["items"][j]["marketID"] then
				local newTimeLeft = BMTSyncDB[syncedRealm]["items"][j]["timeLeft"]
				-- recalculate predictions
				BMT_RecalculateAndSetPredictedTimes(syncedRealm, i, newTimeLeft, currDate, currTime, false)
				-- check if better
				if BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedLow"] > BMTSyncDB[syncedRealm]["items"][j]["predictedLow"] or BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedHigh"] < BMTSyncDB[syncedRealm]["items"][j]["predictedHigh"] then
					-- concatenate this item to the message that will be sent back
					for k=1, #itemShortSyncField do
						if k > 1 then
							betterPredMsg = betterPredMsg .. SEPARATOR
						end
						betterPredMsg = betterPredMsg..BlackMarketTrackerCurrentDB[syncedRealm]["items"][i][itemShortSyncField[k]]
					end
					betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
				end
				break
			end
		end
	end
	-- cut out any unnecessary separators
	if string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR_CATEGORY or string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR then
		betterPredMsg = string.sub(betterPredMsg, 1, string.len(betterPredMsg)-1)
	end
	-- send it back
	BMT:SendCommMessage(ADDON_PREFIX, betterPredMsg, channel, sender)
	-- clean up and start waiting for a the best result after sender puts everything together
	BMTSyncDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS)
end

-- this function just adds received informations
function BMTSyncing_OnBetterResultReceive(categories, syncedRealm, channel, sender)
	if BMTSyncBetterDB[syncedRealm] == nil then
		BMTSyncBetterDB[syncedRealm] = {}
	end
	local index = #BMTSyncBetterDB[syncedRealm]+1
	BMTSyncBetterDB[syncedRealm][index] = {}
	BMTSyncBetterDB[syncedRealm][index]["updateInfo"] = {}
	BMTSyncBetterDB[syncedRealm][index]["items"] = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncBetterDB[syncedRealm][index]["updateInfo"][updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncBetterDB[syncedRealm][index]["updateInfo"][updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	
	--[[ not necessarily true, as we still want the DB that has more accurate predictions, even if it's older than our
	-- ignore searches older than our
	local res = BMT_CompareDateAndTime(BMTSyncBetterDB[syncedRealm][index]["updateInfo"]["updateDate"], BMTSyncBetterDB[syncedRealm][index]["updateInfo"]["updateTime"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateTime"])
	if res == "smaller" or res == "equal" then BMTSyncBetterDB[syncedRealm][index] = nil return end
	]]
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncBetterDB[syncedRealm][index]["items"]
		-- (i-2) so we start from 1
		BMTSyncBetterDB[syncedRealm][index]["items"][(i-2)] = {}
		for j=1, #itemShortSyncField do
			if isItemShortFieldNumber[j] then
				BMTSyncBetterDB[syncedRealm][index]["items"][(i-2)][itemShortSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncBetterDB[syncedRealm][index]["items"][(i-2)][itemShortSyncField[j]] = itemInfo[j]
			end
			if itemShortSyncField[j] == "predictedLow" and BMTSyncBetterDB[syncedRealm][index]["items"][(i-2)][itemShortSyncField[j]] < 0 then
				BMTSyncBetterDB[syncedRealm][index]["items"][(i-2)][itemShortSyncField[j]] = 0
			end
		end
	end
end


function BMTSyncng_OnSearchRequestReceive(categories, syncedRealm, channel, sender)
	-- if there was still something, then remove it
	BMTSyncDB[syncedRealm] = nil
	-- create data struct
	BMTSyncDB[syncedRealm] = {}
	BMTSyncDB[syncedRealm]["updateInfo"] = {}
	BMTSyncDB[syncedRealm]["items"] = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	local haveNewerSearch = false
	-- we only let one know if we have newer search or more accurate predictions OR newer search
	local res = BMT_CompareDateAndTime(BMTSyncDB[syncedRealm]["updateInfo"]["updateDate"], BMTSyncDB[syncedRealm]["updateInfo"]["updateTime"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateTime"])
	if res == "greater"  then 
		haveNewerSearch = true
	end
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncDB[syncedRealm]["items"]
		-- (i-2) so we start from 1
		BMTSyncDB[syncedRealm]["items"][(i-2)] = {}
		for j=1, #itemShortSyncField do
			if isItemShortFieldNumber[j] then
				BMTSyncDB[syncedRealm]["items"][(i-2)][itemShortSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncDB[syncedRealm]["items"][(i-2)][itemShortSyncField[j]] = itemInfo[j]
			end
		end
	end

	local betterPredMsg = BMTSyncing_InitialInfo(KEY_BETTER_PREDICTIONS)
	-- add update info
	for i=1, #updateInfoSyncField do
		if i > 1 then
			betterPredMsg = betterPredMsg .. SEPARATOR
		end
		betterPredMsg = betterPredMsg..BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]]
	end
	betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
	local numItems = BMT_GetNumItems(syncedRealm)
	local currDate = date("%d/%m/%y")
	local currHour, currMinute = GetGameTime()
	local currTime = 60*currHour+currMinute
	-- refresh all predictions before checking anything
	BMT_RefreshAllPredictedTimes(syncedRealm)
	local haveBetterItems = false
	for i=1, numItems do
		for j=1, #BMTSyncDB[syncedRealm]["items"] do
			if BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["marketID"] == BMTSyncDB[syncedRealm]["items"][j]["marketID"] then
				-- check if better
				if BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedLow"] > BMTSyncDB[syncedRealm]["items"][j]["predictedLow"] or BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedHigh"] < BMTSyncDB[syncedRealm]["items"][j]["predictedHigh"] then
					haveBetterItems = true
					-- concatenate this item to the message that will be sent back
					for k=1, #itemShortSyncField do
						if k > 1 then
							betterPredMsg = betterPredMsg .. SEPARATOR
						end
						betterPredMsg = betterPredMsg..BlackMarketTrackerCurrentDB[syncedRealm]["items"][i][itemShortSyncField[k]]
					end
					betterPredMsg = betterPredMsg .. SEPARATOR_CATEGORY
				end
				break
			end
		end
	end
	-- cut out any unnecessary separators
	if string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR_CATEGORY or string.sub(betterPredMsg, string.len(betterPredMsg)) == SEPARATOR then
		betterPredMsg = string.sub(betterPredMsg, 1, string.len(betterPredMsg)-1)
	end
	-- send it back
	if haveNewerSearch == true or haveBetterItems == true then
		-- respond
		BMT:SendCommMessage(ADDON_PREFIX, betterPredMsg, channel, sender)
	end
	-- clean up and start waiting for a the best result after sender puts everything together
	BMTSyncDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS)
end

-- this should contain the newest and most accurate search
-- if the search is older than our - return
-- if any item is less accurate than our, don't write it
function BMTSyncng_OnFullItemDetailsReceive(categories, syncedRealm, channel, sender)
	-- if there was still something, then remove it
	BMTSyncDB[syncedRealm] = nil
	-- create data struct
	BMTSyncDB[syncedRealm] = {}
	BMTSyncDB[syncedRealm]["updateInfo"] = {}
	BMTSyncDB[syncedRealm]["items"] = {}
	local updateInfo = {strsplit(SEPARATOR, categories[2])}
	for i=1, #updateInfoSyncField do
		if isUpdateFieldNumber[i] then
			BMTSyncDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]] = tonumber(updateInfo[i])
		else
			BMTSyncDB[syncedRealm]["updateInfo"][updateInfoSyncField[i]] = updateInfo[i]
		end
	end
	local res = BMT_CompareDateAndTime(BMTSyncDB[syncedRealm]["updateInfo"]["updateDate"], BMTSyncDB[syncedRealm]["updateInfo"]["updateTime"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateTime"])
	if res == "smaller" then return end
	
	-- go through all of the items (items start @ index 3)
	for i=3, #categories do
		local itemInfo = {strsplit(SEPARATOR, categories[i])}
		local numItems = #BMTSyncDB[syncedRealm]["items"]
		-- (i-2) so we start from 1
		BMTSyncDB[syncedRealm]["items"][(i-2)] = {}
		for j=1, #itemFullSyncField do
			if isItemFullSyncFieldNumber[j] then
				BMTSyncDB[syncedRealm]["items"][(i-2)][itemFullSyncField[j]] = tonumber(itemInfo[j])
			else
				BMTSyncDB[syncedRealm]["items"][(i-2)][itemFullSyncField[j]] = itemInfo[j]
			end
		end
	end
	
	
	-- rewrite update info to current
	BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"] = BMTSyncDB[syncedRealm]["updateInfo"]
	
	-- do this just in case
	BMT_MoveAllOldEntriesToHistory(syncedRealm, BMTSyncDB[syncedRealm]["items"], BMTSyncDB[syncedRealm]["updateInfo"]["updateDate"], BMTSyncDB[syncedRealm]["updateInfo"]["updateTime"])
	
	-- refresh predictions before doing anything
	BMT_RefreshAllPredictedTimes(syncedRealm)
	
	local numItems = BMT_GetNumItems(syncedRealm)
	for i=1, numItems do
		for j=1, #BMTSyncDB[syncedRealm]["items"] do
			if BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["marketID"] == BMTSyncDB[syncedRealm]["items"][j]["marketID"] then
				-- check if better (note: this is the other way around than in the rest of the code)
				if BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedLow"] < BMTSyncDB[syncedRealm]["items"][j]["predictedLow"] or BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedHigh"] > BMTSyncDB[syncedRealm]["items"][j]["predictedHigh"] then
					-- rewrite everything to current
					BlackMarketTrackerCurrentDB[syncedRealm]["items"][i] = BMTSyncDB[syncedRealm]["items"][j]
				end
				break
			end
		end
	end
	-- clean up
	BMTSyncDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
	
	-- for bottom text handling only
	BMT_SetNewUpdater(syncedRealm, BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["playerName"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["playerClass"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["playerGuild"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateTime"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateDate"]);
	-- update the list in case the addon's browser is opened
	-- bottom text is already updated in BMT_SetNewUpdater()
	if(BMTFrame:IsShown()) then
		BMTScrollFrame_Update();
	end
end

-- mainly to handle the spinning inset
function BMTSyncing_ChangeSyncingState(newState)
	BMTSyncingState = newState
	if BMTFrame:IsShown() then
		if BMTSyncingState == 0 then
			BMTFrameSpinningInset:Hide()
		else
			BMTFrameSpinningInset:Show()
		end
	end
	if BMTSyncingState == SYNCING_STATE_CHECK_VERSION then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_CHECK_VERSION_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_NEW_SEARCH or BMTSyncingState == SYNCING_STATE_RECEIVING_NEW_SEARCH then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_NEW_SEARCH_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_NARROWING_SEARCH then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_NARROWING_SEARCH_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_AWAITING_FULL_ITEM_DETAILS_TEXT)
	elseif BMTSyncingState == SYNCING_STATE_AWAITING_BETTER_SEARCH then
		BMTFrameSpinningInset.Text:SetText(SYNCING_STATE_AWAITING_BETTER_SEARCH_TEXT)
	end
end

-- returns initial informations about sender that is added to every addon message
-- parameters:
--	- key - key of the sync
-- returns: string in format key%addonVer%realm$
-- where	% is a separator within a group
-- 			$ is a category separator
function BMTSyncing_InitialInfo(key)
	local realmName = GetRealmName()
	local msg = key..SEPARATOR..tostring(BMTRevision)..SEPARATOR..realmName..SEPARATOR_CATEGORY
	
	return msg
end

-- asks for new version everywhere (called only upon logging in)
function BMTSyncing_AskForVersion_OnLogin()
	BMT:RegisterComm(ADDON_PREFIX, "OnCommReceived")
	
	-- no need to check if window is shown here, as this only happens upon logging in -> will be hidden
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_CHECK_VERSION)
	BMTSyncing_BroadcastAddonMessage(KEY_VERREQ..SEPARATOR..BMTRevision)
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_VERSION, BMTSyncing_FinishSyncingVersion)
end

function BMTSyncing_BroadcastCurrentSearch()
	local realmName = GetRealmName()
	local msg = BMTSyncing_InitialInfo(KEY_NEW_SEARCH)
	-- last one has different separator
	for i=1, #updateInfoSyncField do
		if i > 1 then
			msg = msg .. SEPARATOR
		end
		msg = msg..BlackMarketTrackerCurrentDB[realmName]["updateInfo"][updateInfoSyncField[i]]
	end
	msg = msg .. SEPARATOR_CATEGORY
	-- send out short details about each item, so that players can determine wheather some of their items are outdated or not, as well as determine new time predictions for each
	local numItems = BMT_GetNumItems(realmName)
	for j=1, numItems do
		for i=1, #itemShortSyncField do
			if i > 1 then
				msg = msg .. SEPARATOR
			end
			msg = msg..BlackMarketTrackerCurrentDB[realmName]["items"][j][itemShortSyncField[i]]
		end
		if j < numItems then
			msg = msg .. SEPARATOR_CATEGORY
		end
	end
	BMT:SendCommMessage(ADDON_PREFIX, msg, "GUILD")
	-- TODO: Friends, party/raid
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NARROWING_SEARCH)
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_NEW_SEARCH, BMTSyncing_FinishNarrowingSearch, realmName)
end

function BMTSyncing_BroadcastAuctionsRequest()
	if BMTSyncingState ~= SYNCING_STATE_NONE then return end
	
	local realmName = GetRealmName()
	BMT_RefreshAllPredictedTimes(realmName)
	local msg = BMTSyncing_InitialInfo(KEY_BETTER_SEARCH_REQ)
	-- last one has different separator
	for i=1, #updateInfoSyncField do
		if i > 1 then
			msg = msg .. SEPARATOR
		end
		msg = msg..BlackMarketTrackerCurrentDB[realmName]["updateInfo"][updateInfoSyncField[i]]
	end
	msg = msg .. SEPARATOR_CATEGORY
	-- send out short details about each item, so that players can determine wheather some of their items are outdated or not, as well as determine new time predictions for each
	local numItems = BMT_GetNumItems(realmName)
	for j=1, numItems do
		for i=1, #itemShortSyncField do
			if i > 1 then
				msg = msg .. SEPARATOR
			end
			msg = msg..BlackMarketTrackerCurrentDB[realmName]["items"][j][itemShortSyncField[i]]
		end
		if j < numItems then
			msg = msg .. SEPARATOR_CATEGORY
		end
	end
	BMT:SendCommMessage(ADDON_PREFIX, msg, "GUILD")
	-- TODO: Friends, party/raid
	BMTTaskManager_Schedule(SYNCING_TIMEOUT_BETTER_SEARCH_REQ, BMTSyncing_FinishNarrowingSearch, realmName)
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_AWAITING_BETTER_SEARCH)
end

function BMTSyncing_BroadcastFullItemDatabase(realmName)
	if BlackMarketTrackerCurrentDB == nil or BlackMarketTrackerCurrentDB[realmName] == nil then return end
	BMT_RefreshAllPredictedTimes(realmName)
	local msg = BMTSyncing_InitialInfo(KEY_ITEM_FULL)
	-- update info
	for i=1, #updateInfoSyncField do
		if i > 1 then
			msg = msg .. SEPARATOR
		end
		msg = msg..BlackMarketTrackerCurrentDB[realmName]["updateInfo"][updateInfoSyncField[i]]
	end
	msg = msg .. SEPARATOR_CATEGORY
	-- full item detauls
	local numItems = BMT_GetNumItems(realmName)
	for j=1, numItems do
		for i=1, #itemFullSyncField do
			if i > 1 then
				msg = msg .. SEPARATOR
			end
			msg = msg..BlackMarketTrackerCurrentDB[realmName]["items"][j][itemFullSyncField[i]]
		end
		if j < numItems then
			msg = msg .. SEPARATOR_CATEGORY
		end
	end
	BMT:SendCommMessage(ADDON_PREFIX, msg, "GUILD")
	-- TODO: Friends, party/raid
end

---- TIMEOUTS ----
-- bottom text handling, also asks for auctions
function BMTSyncing_FinishSyncingVersion()
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
	-- fade out only if we're not fading already
	if (BMTHighestRevision > BMTRevision and BMTFadeState == 3) and BMTFrame:IsShown() then
		BMTFadeState = -1
		BMTBottomText_FadingHandler()
	end
	BMTSyncing_BroadcastAuctionsRequest()
end

-- this function is called when timeout finishes for both: new search sender, login request sender
function BMTSyncing_FinishNarrowingSearch(syncedRealm)
	-- this has done new search and might've received better predictions, check and set them here
	if BMTSyncingState == SYNCING_STATE_NARROWING_SEARCH and BMTSyncBetterDB ~= nil and BMTSyncBetterDB[syncedRealm] ~= nil then
		-- here we will ignore the update infos completly, as we're just after doing a search, therefore our should be the newest one
		local numBetterDB = #BMTSyncBetterDB[syncedRealm]
		local numItems = #BlackMarketTrackerCurrentDB[syncedRealm]["items"]
		-- won't run if received nothing better
		-- go through all received better results
		for k=1, numBetterDB do
			-- go through all of our items
			for i=1, numItems do
				local numRecvItems = #BMTSyncBetterDB[syncedRealm][k]["items"]
				-- go through all received items from one player
				for j=1, numRecvItems do
					if BMTSyncBetterDB[syncedRealm][k]["items"][j]["marketID"] == BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["marketID"] then
						local foundMarketID = BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["marketID"]
						-- set best to our current
						local bestPredictedLow = BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedLow"]
						local bestPredictedHigh = BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedHigh"]
						-- go through received databases from other players, trying to find better predictions
						for l=1, numBetterDB do
							-- go through all items for each player, if marketID matches - check if prediction is better
							local numRecvPlrItems = #BMTSyncBetterDB[syncedRealm][l]["items"]
							for m=1, numRecvPlrItems do
								if foundMarketID == BMTSyncBetterDB[syncedRealm][l]["items"][m]["marketID"] then
									if BMTSyncBetterDB[syncedRealm][l]["items"][m]["predictedLow"] > bestPredictedLow then
										bestPredictedLow = BMTSyncBetterDB[syncedRealm][l]["items"][m]["predictedLow"]
									end
									if BMTSyncBetterDB[syncedRealm][l]["items"][m]["predictedHigh"] < bestPredictedLow then
										bestPredictedHigh = BMTSyncBetterDB[syncedRealm][l]["items"][m]["predictedHigh"]
									end
									break
								end
							end
						end
						-- here, we should have best predictions for one item, check for weird stuff
						if bestPredictedLow > bestPredictedHigh then break
						else
							BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedLow"] = bestPredictedLow
							BlackMarketTrackerCurrentDB[syncedRealm]["items"][i]["predictedHigh"] = bestPredictedHigh
						end
						break
					end
				end
			end
		end
	-- logged in and asked for newer searches/better responses - maybe got them
	-- if received nothing then don't do anything here
	elseif BMTSyncingState == SYNCING_STATE_AWAITING_BETTER_SEARCH and BMTSyncBetterDB ~= nil and BMTSyncBetterDB[syncedRealm] ~= nil then
		-- newest search isn't necessarily the most accurate on predictions
		-- algorithm is:
		-- a. recalculate predictions to current time (locally only) - because browser could be opened after logging in but before this function got called
		-- b. move all predictions back up again, to the point of our last search
		-- c. get newest search update info
		-- d. remove old auctions, accordingly to items in the newest search
		-- e. recalculate the predictions, as if the search was done at the point of newest search ('fake search')
		-- f. recalculate predictions to current time (locally only)
		
		-- a. recalculate predictions to current time (locally only) - because browser could be opened after logging in but before this function got called
		BMT_RefreshAllPredictedTimes(syncedRealm)
		
		-- b. move all predictions back up again, to the point of our last search
		BMT_MoveAllPredictionsToLastSearchTime(syncedRealm)
		
		-- c. get newest search update info (nil is current)
		local newestUpdateInfoIndex = nil
		local numBetterDB = #BMTSyncBetterDB[syncedRealm]
		for k=1, numBetterDB do
			if newestUpdateInfoIndex == nil then
				-- our is older
				if BMT_CompareDateAndTime(BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[syncedRealm]["updateInfo"]["updateTime"], BMTSyncBetterDB[syncedRealm][k]["updateDate"], BMTSyncBetterDB[syncedRealm][k]["updateTime"]) == "smaller" then
					newestUpdateInfoIndex = k
				end
			else
				-- previous best is older
				if BMT_CompareDateAndTime(BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["updateDate"], BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["updateTime"], BMTSyncBetterDB[syncedRealm][k]["updateDate"], BMTSyncBetterDB[syncedRealm][k]["updateTime"]) == "smaller" then
					newestUpdateInfoIndex = k
				end
			end
		end
		-- if nothing better was found, we'll just send current, after moving predictions back to current 
		
		-- newer search was found
		if newestUpdateInfoIndex ~= nil then
			-- d. remove old auctions, accordingly to items in the newest search
			BMT_MoveAllOldEntriesToHistory(syncedRealm, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["items"], BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["updateInfo"]["updateDate"], BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["updateInfo"]["updateTime"])
			
			-- e. recalculate the predictions, as if the search was done at the point of newest search ('fake search')
			local numItems = BMT_GetNumItems(syncedRealm)
			for i=1, numItems do
				local newTimeLeft = BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["items"][i]["timeLeft"]
				BMT_RecalculateAndSetPredictedTimes(syncedRealm, i, newTimeLeft, BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["updateInfo"]["updateDate"], BMTSyncBetterDB[syncedRealm][newestUpdateInfoIndex]["updateInfo"]["updateTime"], false)
			end
		end
		-- f. recalculate predictions to current time (locally only)
		BMT_RefreshAllPredictedTimes(syncedRealm)
	end
	
	BMTSyncing_BroadcastFullItemDatabase(syncedRealm)
	BMTSyncBetterDB[syncedRealm] = nil
	BMTSyncing_ChangeSyncingState(SYNCING_STATE_NONE)
end


---- SPINNING TILES ----
function BMTSpinningInset_Initialize()
	-- if you change it here, change it in the BlackMarketTracker.xml
	local self = BMTFrameSpinningInset
	-- set rotations
	for i=0, 7 do
		_G[self:GetName().."Tile"..i]:SetRotation(BMTDegToRad(i*45));
		-- original size is 32x32, which is a little to big in this case
		_G[self:GetName().."Tile"..i]:SetWidth(SPINNING_TILE_SIZE);
		_G[self:GetName().."Tile"..i]:SetHeight(SPINNING_TILE_SIZE);
	end
	-- hide all tiles
	for i=0, 7 do
		_G[self:GetName().."Tile"..i]:SetAlpha(0);
	end
	-- fade in first tile
	BMTTileFadeIn(_G[self:GetName().."Tile0"], 0);
end

function BMTDegToRad(degrees)
	return ((PI/180)*degrees);
end

function BMTFadeInTile(frame, id)
	local fadeInfo={
    mode = "IN",
    timeToFade = SPINNING_TILE_FADE_IN_TIME,
    finishedFunc = BMTTileFadeIn,
    finishedArg1 = frame,
	finishedArg2 = id,
    }
    UIFrameFade(frame, fadeInfo);
end

function BMTTileFadeIn(self, id)
	local fadeInfo={
    mode = "OUT",
    timeToFade = SPINNING_TILE_FADE_OUT_TIME,
    finishedFunc = BMTTileFadeOut,
    finishedArg1 = self,
	finishedArg2 = id,
    }
    UIFrameFade(self, fadeInfo);

	-- fade in next too
	local nextId, nextTile;
	if (id>0) then
		nextId = id-1;
	else
		nextId = 7;
	end
	
	nextTile = _G[self:GetParent():GetName().."Tile"..nextId];
	BMTFadeInTile(nextTile, nextId);
end

-- for future use, if ever needed
function BMTTileFadeOut(self, id)	
end
